/////////////////////////////////////////////////////////////////////////////
///
/// @file datavector.h
///
/// Data conversion for sending/receiving: std::vectors and hashed sets
/// of any objects.
/// This file contains the definitions of the operators << and >>
/// for encoding some variables to the binary data structure
/// and retrieving them. This binary data structure is used
/// in the communication between the coordinator and workers.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2013 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on February 15, 2008. Last revision: March 26, 2013.


#ifndef _DATAVECTOR_H_
#define _DATAVECTOR_H_


// include some standard C++ header files
#include <vector>
#include <utility>

// include selected header files from the CHomP library
#include "chomp/system/config.h"
#include "chomp/system/textfile.h"
#include "chomp/struct/hashsets.h"
#include "chomp/multiwork/mw.h"


// --------------------------------------------------
// ------------- vector of any objects --------------
// --------------------------------------------------

/// Writes a standard vector of any objects to a MultiWork data structure.
template<class AnyType>
inline chomp::multiwork::mwData &operator << (chomp::multiwork::mwData &data,
	const std::vector<AnyType> &vect)
{
	size_t size = vect. size ();
	data << size;
	for (size_t i = 0; i < size; ++ i)
		data << vect [i];
	return data;
} /* operator << */

/// Reads a standard vector of any objects from a MultiWork data structure.
template<class AnyType>
inline chomp::multiwork::mwData &operator >> (chomp::multiwork::mwData &data,
	std::vector<AnyType> &vect)
{
	size_t size = 0;
	data >> size;
	if (size > 0)
		vect. resize (size);
	else
		vect. clear ();
	for (size_t i = 0; i < size; ++ i)
		data >> vect [i];
	return data;
} /* operator >> */


// --------------------------------------------------
// -------------- pair of any objects ---------------
// --------------------------------------------------

/// Writes a standard pair of any objects to a MultiWork data structure.
template<class Type1, class Type2>
inline chomp::multiwork::mwData &operator << (chomp::multiwork::mwData &data,
	const std::pair<Type1,Type2> &p)
{
	data << p. first;
	data << p. second;
	return data;
} /* operator << */

/// Reads a standard pair of any objects from a MultiWork data structure.
template<class Type1, class Type2>
inline chomp::multiwork::mwData &operator >> (chomp::multiwork::mwData &data,
	std::pair<Type1,Type2> &p)
{
	data >> p. first;
	data >> p. second;
	return data;
} /* operator >> */


// --------------------------------------------------
// ----------- hashed set of any objects ------------
// --------------------------------------------------

/// Writes a standard vector of any objects to a MultiWork data structure.
template<class AnyType>
inline chomp::multiwork::mwData &operator << (chomp::multiwork::mwData &data,
	const chomp::homology::hashedset<AnyType> &vect)
{
	int_t size = vect. size ();
	data << size;
	for (int_t i = 0; i < size; ++ i)
		data << vect [i];
	return data;
} /* operator << */

/// Reads a standard vector of any objects from a MultiWork data structure.
template<class AnyType>
inline chomp::multiwork::mwData &operator >> (chomp::multiwork::mwData &data,
	chomp::homology::hashedset<AnyType> &vect)
{
	int_t size = 0;
	data >> size;
	for (int_t i = 0; i < size; ++ i)
	{
		AnyType object;
		data >> object;
		vect. add (object);
	}
	return data;
} /* operator >> */


#endif // _DATAVECTOR_H_

