/////////////////////////////////////////////////////////////////////////////
///
/// @file expdata.h
///
/// A data structure for gathering, saving and retrieving expansion data.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2006-2013 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on August 8, 2013. Last revision: August 13, 2014.


#ifndef _expdata_h_
#define _expdata_h_


// standard library headers
#include <string>
#include <vector>
#include <iostream>
#include <new>
#include <memory>
//#define NDEBUG
#include <assert.h>

// selected header files from the CHomP library
#include "chomp/system/config.h"
#include "chomp/system/textfile.h"
//#include "chomp/system/timeused.h"


// --------------------------------------------------
// ---------------------- min3 ----------------------
// --------------------------------------------------

/// Computes the minimum of the three elements.
template <class T>
inline const T&min3 (const T &x, const T &y, const T &z)
{
	return ((x < y) ? ((x < z) ? x : z) : ((y < z) ? y : z));
} /* min3 */


// --------------------------------------------------
// ----------------- expansion data -----------------
// --------------------------------------------------

/// A data structure for storing expansion data for subintervals
/// of a larger interval.
class ExpansionData
{
public:
	/// The default constructor.
	/// No destructor is necessary.
	/// The default copy constructor and assignment operator is fine.
	ExpansionData ();

	/// Sets the interval and the number of subintervals.
	void initialize (const double &left, const double &right,
		int_t nIntervals);

	/// Returns the left bound of the main interval.
	const double &leftBound () const;

	/// Returns the right bound of the main interval.
	const double &rightBound () const;

	/// Returns the number of subintervals.
	int_t size () const;

	/// Sets the data for the specific subinterval.
	void set (int_t n, const double &delta, const double &lambdaCycles,
		const double &lambdaStart, const double &lambdaEnd);

	/// Retrieves the data for the specific subinterval.
	void get (int_t n, double &delta, double &lambdaCycles,
		double &lambdaStart, double &lambdaEnd) const;

	/// Writes the header of a data collection to a given stream.
	/// Sets the size of integers if it was not yet fixed.
	void writeHeader (std::ostream &out);

	/// Reads the header of a data collection from a given stream.
	/// Initializes the data structure if not yet initialized,
	/// or checks the compatibility of the data.
	/// Returns 0 if successful, -1 if the data is incompatible
	/// (e.g. the numbers were saved from a different architecture).
	int readHeader (std::istream &in);

	/// Writes the data to a given stream.
	/// If requested, does not include the data previously read,
	/// and saves newly added data only.
	/// The stream must be opened in the binary mode.
	/// If the beginning of the stream must contain the header
	/// then the header must be written before writing the data.
	/// Returns the number of data objects written to the stream.
	int_t writeData (std::ostream &out) const;

	/// Reads data from a given stream.
	/// The stream must be opened in the binary mode.
	/// If there is any header in the stream, it must be read
	/// before reading the data.
	/// Returns the number of data objects read from the file.
	int_t readData (std::istream &in);

	/// Sets a stream to which all the data will be saved
	/// while being added to the data collection (NULL for none).
	/// The stream must be opened in the binary mode.
	/// If a header is necessary at the beginning of the stream,
	/// it must be written separately.
	void setStream (std::ostream *out);

	/// Returns the pointer to the stream to which the data
	/// is being saved or NULL if none.
	std::ostream *getStream ();

	/// Should the data stream be flushed after every data piece?
	bool flush;

private:
	/// The left bound of the entire interval.
	double m_left;

	/// The right bound of the entire interval.
	double m_right;

	/// The number of intervals in the subdivision.
	int_t m_nIntervals;

	/// The computed minimal delta > 0 for which lambda is positive.
	/// Special values: 0 = not computed, -1 = no positive lambda found.
	std::vector<double> m_delta;

	/// The computed values of lambda for cycles for the minimal delta.
	/// Special values: 0 = not computed, -1 = no positive lambda found.
	std::vector<double> m_lambdaCycles;

	/// The computed values of lambda for starting paths for the delta.
	/// Special values: 0 = not computed, -1 = no positive lambda found.
	std::vector<double> m_lambdaStart;

	/// The computed values of lambda for ending paths for the delta.
	/// Special values: 0 = not computed, -1 = no positive lambda found.
	std::vector<double> m_lambdaEnd;

	/// A stream to which newly added results of computations
	/// should be written.
	/// Note that the stream should be opened in the binary mode.
	std::ostream *m_stream;

	/// The size of the (integer) index in the data in bytes.
	/// The value of 0 means that no data was yet written or read.
	int m_intSize;

	/// Writes an integer of the given length to a binary stream.
	static void writeInt (std::ostream &out, int_t value, int length);

	/// Reads an integer of the given length from a binary stream.
	static int_t readInt (std::istream &in, int length);

	/// Writes a double-precision number to a binary stream.
	static void writeDouble (std::ostream &out, double number);

	/// Reads a double-precision number from a binary stream.
	static double readDouble (std::istream &in);

}; /* class ExpansionData */

// --------------------------------------------------

inline ExpansionData::ExpansionData ():
	flush (true), m_left (0), m_right (0), m_nIntervals (0),
	m_stream (0), m_intSize (0)
{
	return;
} /* ExpansionData::ExpansionData */

inline void ExpansionData::initialize (const double &left,
	const double &right, int_t nIntervals)
{
	if (left >= right)
		throw "Wrong bounds requested for ExpansionData.";
	if (nIntervals <= 0)
		throw "Non-positive number of intervals in ExpansionData.";
	m_left = left;
	m_right = right;
	m_nIntervals = nIntervals;
	if (!m_delta. empty ())
		std::vector<double> (). swap (m_delta);
	m_delta. resize (nIntervals, 0);
	if (!m_lambdaCycles. empty ())
		std::vector<double> (). swap (m_lambdaCycles);
	m_lambdaCycles. resize (nIntervals, 0);
	if (!m_lambdaStart. empty ())
		std::vector<double> (). swap (m_lambdaStart);
	m_lambdaStart. resize (nIntervals, 0);
	if (!m_lambdaEnd. empty ())
		std::vector<double> (). swap (m_lambdaEnd);
	m_lambdaEnd. resize (nIntervals, 0);
	return;
} /* ExpansionData::initialize */

inline const double &ExpansionData::leftBound () const
{
	return m_left;
} /* ExpansionData::leftBound */

inline const double &ExpansionData::rightBound () const
{
	return m_right;
} /* ExpansionData::rightBound */

inline int_t ExpansionData::size () const
{
	return m_nIntervals;
} /* ExpansionData::size */

inline void ExpansionData::set (int_t n, const double &delta,
	const double &lambdaCycles,
	const double &lambdaStart, const double &lambdaEnd)
{
	if ((n < 0) || (n >= m_nIntervals))
		throw "Index out of range in ExpansionData (set).";
	m_delta [n] = delta;
	m_lambdaCycles [n] = lambdaCycles;
	m_lambdaStart [n] = lambdaStart;
	m_lambdaEnd [n] = lambdaEnd;
//	chomp::homology::sbug << "DEBUG: ExpD.set (" <<
//		n << ", " << delta << ", " << lambdaCycles << ", " <<
//		lambdaStart <<  ", " << lambdaEnd << ").\n";
	if (m_stream)
	{
		if (!m_intSize)
			m_intSize = sizeof (int_t);
		writeInt (*m_stream, n, m_intSize);
		writeDouble (*m_stream, delta);
		writeDouble (*m_stream, lambdaCycles);
		writeDouble (*m_stream, lambdaStart);
		writeDouble (*m_stream, lambdaEnd);
		if (flush)
			m_stream -> flush ();
	}
	return;
} /* ExpansionData::set */

inline void ExpansionData::get (int_t n, double &delta, double &lambdaCycles,
	double &lambdaStart, double &lambdaEnd) const
{
	if ((n < 0) || (n >= m_nIntervals))
		throw "Index out of range in ExpansionData (get).";
	delta = m_delta [n];
	lambdaCycles = m_lambdaCycles [n];
	lambdaStart = m_lambdaStart [n];
	lambdaEnd = m_lambdaEnd [n];
	return;
} /* ExpansionData::get */

inline void ExpansionData::writeHeader (std::ostream &out)
{
	if (!m_nIntervals)
	{
		throw "Trying to write the header of uninitialized "
			"ExpansionData.";
	}
	if (!m_intSize)
		m_intSize = sizeof (int_t);
	const char *mark = "BinData:\n\n";
	for (int i = 0; mark [i]; ++ i)
		out. put (mark [i]);
	out. put (sizeof (double));
	out. put (m_intSize);
	const double testNumber = 1800.1900e4;
	writeDouble (out, testNumber);
	writeDouble (out, m_left);
	writeDouble (out, m_right);
	writeInt (out, m_nIntervals, m_intSize);
	return;
} /* ExpansionData::writeHeader */

inline int ExpansionData::readHeader (std::istream &in)
{
	// read the header
	const char *mark = "BinData:\n\n";
	for (int i = 0; mark [i]; ++ i)
	{
		if (in. get () != mark [i])
			return -1;;
	}
	if (in. get () != sizeof (double))
		return -1;
	int intSize = in. get ();
	if (m_intSize && (m_intSize != intSize))
		return -1;
	m_intSize = intSize;
	const double testNumber = 1800.1900e4;
	if (readDouble (in) != testNumber)
		return -1;
	double left = readDouble (in);
	double right = readDouble (in);
	int nIntervals = readInt (in, m_intSize);

	// check the compatibility of the data with the existing data
	if (m_nIntervals && (nIntervals != m_nIntervals))
		return -1;
	if ((m_left || m_right) && ((left != m_left) || (right != m_right)))
		return -1;

	// initialize the data structure if necessary
	if (!m_nIntervals)
		initialize (left, right, nIntervals);

	return 0;
} /* ExpansionData::readHeader */

inline int_t ExpansionData::writeData (std::ostream &out) const
{
	int_t count (0);
	for (int n = 0; n < m_nIntervals; ++ n)
	{
		if (!m_delta [n])
			continue;
		writeInt (out, n, m_intSize);
		writeDouble (out, m_delta [n]);
		writeDouble (out, m_lambdaCycles [n]);
		writeDouble (out, m_lambdaStart [n]);
		writeDouble (out, m_lambdaEnd [n]);
		++ count;
	}
	return count;
} /* ExpansionData::writeData */

inline int_t ExpansionData::readData (std::istream &in)
{
	int_t count (0);
	while (in. peek () != EOF)
	{
		int n = readInt (in, m_intSize);
		double delta = readDouble (in);
		double lambdaCycles = readDouble (in);
		double lambdaStart = readDouble (in);
		double lambdaEnd = readDouble (in);
		set (n, delta, lambdaCycles, lambdaStart, lambdaEnd);
		++ count;
	}
	return count;
} /* ExpansionData::readData */

inline void ExpansionData::setStream (std::ostream *out)
{
	m_stream = out;
	return;
} /* ExpansionData::setStream */

inline std::ostream *ExpansionData::getStream ()
{
	return m_stream;
} /* ExpansionData::getStream */

// --------------------------------------------------

inline void ExpansionData::writeInt (std::ostream &out, int_t value,
	int length)
{
	int shift = 0;
	for (int i = 0; i < length; ++ i)
	{
		out. put ((value >> shift) & 0xFF);
		shift += 8;
	}
	return;
} /* ExpansionData::writeInt */

inline int_t ExpansionData::readInt (std::istream &in, int length)
{
	int_t value (0);
	int shift = 0;
	for (int i = 0; i < length; ++ i)
	{
		if (in. peek () == EOF)
			throw "Input file ended unexpectedly.";
		int_t byte = in. get ();
		value |= (byte << shift);
		shift += 8;
	}
	return value;
} /* ExpansionData::readInt */

inline void ExpansionData::writeDouble (std::ostream &out, double number)
{
	unsigned char *bytes = reinterpret_cast<unsigned char *> (&number);
	int size = sizeof (double);
	for (int i = 0; i < size; ++ i)
		out. put (bytes [i]);
	return;
} /* ExpansionData::writeDouble */

inline double ExpansionData::readDouble (std::istream &in)
{
	double number (0);
	unsigned char *bytes = reinterpret_cast<unsigned char *> (&number);
	int size = sizeof (double);
	for (int i = 0; i < size; ++ i)
	{
		if (in. peek () == EOF)
			throw "Input file ended unexpectedly.";
		bytes [i] = in. get ();
	}
	return number;
} /* ExpansionData::readDouble */

// --------------------------------------------------

/// Determines the best delta and the corresponding expansion lambda
/// for the given parameter interval, based on the collection of
/// tabulated expansion data blocks.
/// Returns true if positive values found, false otherwise.
inline bool findBestExpansion
	(const std::vector<ExpansionData *> &expDataBlocks,
	const double &left, const double &right,
	double &delta, double &lambdaCycles,
	double &lambdaStart, double &lambdaEnd)
{
	using chomp::homology::sbug;

	// remember the best values found so far
	double bestDelta = 0;
	double bestLambdaCycles = 0;
	double bestLambdaStart = 0;
	double bestLambdaEnd = 0;

//	sbug << "DEBUG: Looking up expansion data for [" << left <<
//		", " << right << "].\n";

	// go through all the expansion data blocks and gather the results
	for (std::vector<ExpansionData *>::const_iterator it =
		expDataBlocks. begin (); it != expDataBlocks. end (); ++ it)
	{
		// determine the left and right bound of the entire interval
		// covered in the expansion data block, and the number
		// of intervals into which the entire interval is subdivided
		const ExpansionData &expData = *(*it);
		const double leftBound = expData. leftBound ();
		const double rightBound = expData. rightBound ();
		const int_t nIntervals = expData. size ();

		// if the interval is out of scope then skip the block
		if ((left < leftBound) || (right > rightBound))
			continue;

		// compute approximate left and right endpoints
		int_t leftInt = (left == leftBound) ? 0 :
			static_cast<int_t> ((left - leftBound) /
			(rightBound - leftBound) * nIntervals);
		int_t rightInt = (right == rightBound) ? nIntervals :
			static_cast<int_t> ((right - leftBound) /
			(rightBound - leftBound) * nIntervals);

	//	sbug << "DEBUG: leftInt = " << leftInt << ", rightInt = " <<
	//		rightInt << ".\n";
	//	sbug << "DEBUG: leftBound = " << leftBound <<
	//		", rightBound = " << rightBound << ".\n";

		// do corrections to make the endpoints rigorous
		int corrections = 0;
		if (left != leftBound)
		{
			while (leftBound + leftInt *
				(rightBound - leftBound) / nIntervals > left)
			{
				-- leftInt;
				++ corrections;
			}
			while (leftBound + (leftInt + 1) *
				(rightBound - leftBound) / nIntervals <= left)
			{
				++ leftInt;
				++ corrections;
			}
		}
		if (right != rightBound)
		{
			while (leftBound + rightInt *
				(rightBound - leftBound) / nIntervals < right)
			{
				++ rightInt;
				++ corrections;
			}
			while (leftBound + (rightInt - 1) *
				(rightBound - leftBound) / nIntervals >= right)
			{
				-- rightInt;
				++ corrections;
			}
		}

	//	sbug << "DEBUG: leftInt = " << leftInt << ", rightInt = " <<
	//		rightInt << ".\n";

		// if the intervals for which the expansion data is computed
		// are too fine then skip this information
		// TODO: Is it possible to use min or max delta or lambda???
		if (rightInt - leftInt != 1)
			continue;

		// check if better delta and lambda have been obtained
		double localDelta (0);
		double localLambdaCycles (0);
		double localLambdaStart (0);
		double localLambdaEnd (0);
		expData. get (leftInt, localDelta, localLambdaCycles,
			localLambdaStart, localLambdaEnd);
	//	sbug << "DEBUG: delta = " << localDelta <<
	//		", lambdaCycles = " << localLambdaCycles <<
	//		", lambdaStart = " << localLambdaStart <<
	//		", lambdaEnd = " << localLambdaEnd << ".\n";
		if ((!bestDelta && localDelta) || ((localDelta > 0) &&
			((bestDelta < 0) || (bestDelta > localDelta))))
		{
			bestDelta = localDelta;
			bestLambdaCycles = localLambdaCycles;
			bestLambdaStart = localLambdaStart;
			bestLambdaEnd = localLambdaEnd;
		}
	}

	// return the optimal result if found
	if (bestDelta)
	{
		delta = bestDelta;
		lambdaCycles = bestLambdaCycles;
		lambdaStart = bestLambdaStart;
		lambdaEnd = bestLambdaEnd;
		return true;
	}
	else
		return false;
} /* findBestExpansion */


#endif // _expdata_h_

