/// @addtogroup unifexp
/// @{

/////////////////////////////////////////////////////////////////////////////
///
/// @file maptype.h
///
/// This file contains the definition of an abstract map type.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2007 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on March 8, 2007. Last revision: June 13, 2013.


#ifndef _maptype_h_
#define _maptype_h_

#include <vector>
#include <string>
#include <iostream>


namespace unifexp {

// --------------------------------------------------
// -------------------- map type --------------------
// --------------------------------------------------

/// This is an abstract map type.
/// Each specific map class must inherit from this class.
template <class numType>
class mapType
{
public:
	/// The constructor.
	mapType ();

	/// The virtual destructor.
	virtual ~mapType ();

	/// Allocates a new object of the same type and returns its pointer.
	/// Does not set any parameters, though.
	virtual mapType<numType> *clone () const;

	/// Returns the name of the object.
	virtual std::string name () const = 0;

	/// Sets the parameters of the map.
	void setParam (const numType &_paramMin, const numType &_paramMax);

	/// Returns the left parameter bound of the map.
	const numType &getMinParam () const;

	/// Returns the right parameter bound of the map.
	const numType &getMaxParam () const;

	/// Returns the number of critical points.
	virtual int countCritical () const = 0;

	/// Returns the subsequent critical points.
	virtual numType criticalPoint (int n) const = 0;

	/// Returns the left bound of the domain of the map.
	virtual numType leftBound () const = 0;

	/// Returns the right bound of the domain of the map.
	virtual numType rightBound () const = 0;

	/// Computes an enclosure of the image of the given interval
	/// by the map over the entire interval of parameters.
	virtual void image (const numType &x1, const numType &x2,
		numType &y1, numType &y2) const = 0;

	/// Computes an enclosure of the image of the given interval
	/// by the derivarive of the map.
	virtual void derivative (const numType &x1, const numType &x2,
		numType &y1, numType &y2) const;

	/// Computes the minimal log of the derivative over those points
	/// in the interval [x1,x2] whose images may fall into [y1,y2]
	virtual numType minLogDerivative
		(const numType &x1, const numType &x2,
		const numType &y1, const numType &y2) const = 0;

protected:
	/// The minimal parameter of the map.
	numType paramMin;

	/// The maximal parameter of the map.
	numType paramMax;

private:
	/// The copy constructor is not allowed.
	mapType (const mapType<numType> &);

	/// The assignment operator is not allowed.
	mapType<numType> &operator = (const mapType<numType> &);

}; /* class mapType */

// --------------------------------------------------

template <class numType>
inline mapType<numType>::mapType ()
{
	return;
} /* mapType::mapType */

template <class numType>
inline mapType<numType>::~mapType ()
{
	return;
} /* mapType::~mapType */

template <class numType>
inline mapType<numType> *mapType<numType>::clone () const
{
	throw "Cloning of the map object not defined.";
} /* mapType::clone */

template <class numType>
inline mapType<numType>::mapType (const mapType<numType> &)
{
	throw "Copy constructor not implemented for map types.";
	return;
} /* mapType::mapType */

template <class numType>
inline mapType<numType> &mapType<numType>::operator =
	(const mapType<numType> &)
{
	throw "Assignment operator not implemented for map types.";
	return *this;
} /* mapType::operator = */

template <class numType>
inline void mapType<numType>::setParam (const numType &_paramMin,
	const numType &_paramMax)
{
	paramMin = _paramMin;
	paramMax = _paramMax;
	if (_paramMin > _paramMax)
	{
		paramMin = _paramMax;
		paramMax = _paramMin;
	}
	return;
} /* mapType::setParam */

template <class numType>
inline const numType &mapType<numType>::getMinParam () const
{
	return paramMin;
} /* mapType::getMinParam */

template <class numType>
inline const numType &mapType<numType>::getMaxParam () const
{
	return paramMax;
} /* mapType::getMaxParam */

template <class numType>
inline void mapType<numType>::derivative
	(const numType &, const numType &, numType &, numType &) const
{
	throw "Derivative of the map not defined.";
} /* mapType::derivative */


} // namespace unifexp

#endif // _maptype_h_

/// @}

