/////////////////////////////////////////////////////////////////////////////
///
/// @file mpfrsupp.h
///
/// Supplementary functions and classes for using the MPFR class.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2008 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on March 27, 2007. Last revision: August 16, 2013.


#ifndef _mpfrsupp_h_
#define _mpfrsupp_h_


// standard library headers
#include <string>
#include <sstream>
#include <new>

// headers for multi-precision rigorous arithmetic
#include <gmp.h>
#include <mpfr.h>


// --------------------------------------------------
// ---- functions missing in older MPFR versions ----
// --------------------------------------------------

#if (MPFR_VERSION_MAJOR < 3)

inline void mpfr_set_zero (mpfr_t x, int /*sign*/)
{
	mpfr_set_si (x, 0, GMP_RNDN);
	return;
} /* mpfr_set_zero */

#endif


// --------------------------------------------------
// ----- automatic initialization and clearing ------
// --------------------------------------------------

/// A class whose objects are MPFR numbers which are automatically
/// initialized at initialization and cleared at destruction.
class auto_mpfr
{
public:
	/// The constructor.
	auto_mpfr (int precision = 0);

	/// The destructor.
	~auto_mpfr ();

	/// The conversion to a pointer to the mpfr number.
	operator mpfr_ptr ();

	/// The conversion to a pointer to the const mpfr number.
	operator mpfr_srcptr () const;

	/// The extraction of the actual structure types.
	mpfr_ptr operator -> ();

	/// The extraction of the actual structure types.
	mpfr_srcptr operator -> () const;

	/// Swaps the two numbers.
	auto_mpfr &swap (auto_mpfr &other);

	/// Sets default precision.
	static void setDefaultPrecision (int precision);

private:
	/// The number stored in the class.
	mpfr_t number;

	/// The default precision.
	static int defaultPrecision;

	/// The copy constructor is not allowed.
	auto_mpfr (const auto_mpfr &) {return;}

	/// The assignment operator is not allowed.
	auto_mpfr &operator = (const auto_mpfr &) {return *this;}

}; /* class auto_mpfr */

// --------------------------------------------------

inline auto_mpfr::auto_mpfr (int precision)
{
	int prec = (precision > 0) ? precision : defaultPrecision;
	if (prec <= 0)
		throw "Default precision of auto_mpfr not initialized.";
	mpfr_init2 (number, prec);
	return;
} /* auto_mpfr::auto_mpfr */

inline auto_mpfr::~auto_mpfr ()
{
	mpfr_clear (number);
	return;
} /* auto_mpfr::~auto_mpfr */

inline auto_mpfr::operator mpfr_ptr ()
{
	return number;
} /* auto_mpfr::operator mpfr_ptr */

inline auto_mpfr::operator mpfr_srcptr () const
{
	return number;
} /* auto_mpfr::operator mpfr_srcptr */

inline mpfr_ptr auto_mpfr::operator -> ()
{
	return number;
} /* auto_mpfr::operator -> */

inline mpfr_srcptr auto_mpfr::operator -> () const
{
	return number;
} /* auto_mpfr::operator -> */

inline auto_mpfr &auto_mpfr::swap (auto_mpfr &other)
{
	mpfr_swap (number, other. number);
	return *this;
} /* auto_mpfr::swap */

inline void auto_mpfr::setDefaultPrecision (int precision)
{
	defaultPrecision = precision;
	return;
} /* auto_mpfr::setDefaultPrecision */


// --------------------------------------------------
// ------------- conversion to a string -------------
// --------------------------------------------------

inline std::string mpfr2string (const auto_mpfr &number, int ndigits,
	const mp_rnd_t &rounding)
{
	// determine the length of the buffer
	int length = ndigits + 2;
	if (length < 7)
		length = 7;

	// prepare a temporary buffer
	char *buffer = new char [length];

	// decode the number to the text format
	int base = 10;
	mp_exp_t exponent = 0;
	mpfr_get_str (buffer, &exponent, base, ndigits,
		number, rounding);

	// prepare a human-readable string
	std::ostringstream str;
	str << buffer [0];
	if (buffer [0] == '-')
		str << buffer [1] << '.' << (buffer + 2);
	else
		str << '.' << (buffer + 1);
	if ((buffer [0] == '0') && (exponent == 0));
	else if (exponent != 1)
		str << 'e' << (exponent - 1);
	delete [] buffer;

	// return the resulting string
	return str. str ();

} /* mpfr2string */

inline std::string mpfr2stringDown (const auto_mpfr &number)
{
	char str [128];
	mpfr_sprintf (str, "%.20RDg", static_cast<mpfr_srcptr> (number));
	return std::string (str);
} /* mpfr2stringDown */

inline std::string mpfr2string (const auto_mpfr &number)
{
	char str [128];
	mpfr_sprintf (str, "%.20RNg", static_cast<mpfr_srcptr> (number));
	return std::string (str);
} /* mpfr2string */


#endif // _mpfrsupp_h_

