/////////////////////////////////////////////////////////////////////////////
///
/// @file numdouble.h
///
/// Numerical operations with correct rounding, using the built-in
/// double precision numbers.
/// The operations are programmed using the boost library where possible,
/// and with the CAPD interval arithmetic package where otherwise.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2006-2013 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on June 17, 2013. Last revision: October 13, 2013.


#ifndef _numdouble_h_
#define _numdouble_h_

// local header files
#include "intvcapd.h"
#include "rounding.h"

// standard library headers
//#define NDEBUG
//#include <assert.h>

// selected header files from the CHomP library
//#include "chomp/system/config.h"
//#include "chomp/system/textfile.h"


// --------------------------------------------------
// ---------------- basic operations ----------------
// --------------------------------------------------

/// A namespace with a bunch of rigorous numeric operations.
namespace numdouble
{
/// The type of a floating-point number.
typedef double NumberType;

// --------------------------------------------------

/// The type of an interval of floating-point numbers.
typedef unifexp::IntervalType IntervalType;

using unifexp::intervalExp;
using unifexp::intervalPower;

/// Extracts the left bound of a number (i.e. returns the number).
inline const NumberType &get_left_bound (const NumberType &x)
{
	return x;
} /* get_left_bound */

/// Extracts the left bound of an interval.
inline NumberType get_left_bound (const IntervalType &x)
{
	return x. leftBound ();
} /* get_left_bound */

/// Extracts the left bound of a number (i.e. returns the number).
inline const NumberType &get_right_bound (const NumberType &x)
{
	return x;
} /* get_right_bound */

/// Extracts the right bound of an interval.
inline NumberType get_right_bound (const IntervalType &x)
{
	return x. rightBound ();
} /* get_right_bound */

// --------------------------------------------------


/*
template <class Type1, class Type2>
inline void add_down (NumberType &result, const Type1 &x, const Type2 &y)
{
	unifexp::tRounding<NumberType> rounding;
	result = rounding. WHAT ## _down
		(get_left_bound (x), get_left_bound (y));
	return;
}
*/

/// The basic two-argument operations with correct rounding,
/// defined using the BOOST library.
#define DEFOP(WHAT) \
	inline void WHAT ## _down (NumberType &result, \
		const NumberType &x, const NumberType &y) \
	{ \
		unifexp::tRounding<NumberType> rounding; \
		result = rounding. WHAT ## _down (x, y); \
		return; \
	} \
	\
	inline void WHAT ## _up (NumberType &result, \
		const NumberType &x, const NumberType &y) \
	{ \
		unifexp::tRounding<NumberType> rounding; \
		result = rounding. WHAT ## _up (x, y); \
		return; \
	} \
	\
	inline void WHAT ## _both (IntervalType &result, \
		const NumberType &x, const NumberType &y) \
	{ \
		unifexp::tRounding<NumberType> rounding; \
		result = IntervalType (rounding. WHAT ## _down (x, y), \
			rounding. WHAT ## _up (x, y)); \
		return; \
	}
DEFOP(add)
DEFOP(sub)
DEFOP(mul)
DEFOP(div)
#undef DEFOP

/// A few basic single-argument functions with correct rounding,
/// defined using the CAPD library.
#define DEFOP(WHAT1,WHAT2) \
	inline void WHAT1 ## _down (NumberType &result, const NumberType &x) \
	{ \
		result = WHAT2 (IntervalType (x, x)). leftBound (); \
		unifexp::resetRounding (); \
		return; \
	} \
	\
	inline void WHAT1 ## _up (NumberType &result, const NumberType &x) \
	{ \
		result = WHAT2 (IntervalType (x, x)). rightBound (); \
		unifexp::resetRounding (); \
		return; \
	} \
	\
	inline void WHAT1 ## _both (IntervalType &result, const NumberType &x) \
	{ \
		result = WHAT2 (IntervalType (x, x)); \
		unifexp::resetRounding (); \
		return; \
	}
DEFOP(exp,intervalExp)
DEFOP(log,log)
#undef DEFOP

/// A few basic two-argument functions with correct rounding,
/// defined using the CAPD library.
#define DEFOP(WHAT1,WHAT2) \
	inline void WHAT1 ## _down (NumberType &result, \
		const NumberType &x, const NumberType &y) \
	{ \
		result = WHAT2 (IntervalType (x, x), \
			IntervalType (y, y)). leftBound (); \
		unifexp::resetRounding (); \
		return; \
	} \
	\
	inline void WHAT1 ## _up (NumberType &result, \
		const NumberType &x, const NumberType &y) \
	{ \
		result = WHAT2 (IntervalType (x, x), \
			IntervalType (y, y)). rightBound (); \
		unifexp::resetRounding (); \
		return; \
	} \
	\
	inline void WHAT1 ## _both (IntervalType &result, \
		const NumberType &x, const NumberType &y) \
	{ \
		result = WHAT2 (IntervalType (x, x), \
			IntervalType (y, y)); \
		unifexp::resetRounding (); \
		return; \
	}
DEFOP(power,intervalPower)
#undef DEFOP


} /* namespace numdouble */


#endif // _numdouble_h_

