/// @addtogroup unifexp
/// @{

/////////////////////////////////////////////////////////////////////////////
///
/// @file partsegm.h
///
/// This file contains the definition of a segmented partition type in which
/// each component of the complement of the critical neighborhood
/// is divided into the same number of intervals.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2007 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on March 8, 2007. Last revision: November 6, 2013.

#ifndef _partsegm_h_
#define _partsegm_h_

#include "rounding.h" // (this header must be included first - BOOST)
#include <vector>
#include <string>
#include "parttype.h"


namespace unifexp {

// --------------------------------------------------
// --------------- uniform partition ----------------
// --------------------------------------------------

/// A segmented partition type.
/// Each component of the complement of the critical neighborhood
/// is divided into the same number of intervals.
template <class numType>
class partSegmented: public partType<numType>
{
public:
	/// The constructor which optionally takes a specific name.
	partSegmented (const std::string &name = std::string ("segmented"));

	/// Creates a partition based on the given map, the requested
	/// number of elements in the partition, and the width of the
	/// critical neighborhood.
	void create (const mapType<numType> &theMap, int partCount,
		const numType &delta);

private:
	/// Fills part of the partition table between the given entries
	/// in a uniform way based on the values at these ends.
	void fillUniform (int first, int last);

}; /* class partSegmented */

// --------------------------------------------------

template <class numType>
inline partSegmented<numType>::partSegmented (const std::string &name):
	partType<numType> (name)
{
	return;
} /* partSegmented::partSegmented */

template <class numType>
void partSegmented<numType>::fillUniform (int first, int last)
{
	const numType &numFirst = (*this) [first];
	const numType &numLast = (*this) [last];
	const numType numDiff = (numLast - numFirst) / (last - first);
	for (int i = first + 1; i < last; ++ i)
		(*this) [i] = numFirst + (i - first) * numDiff;
	return;
} /* partSegmented::fillUniform */

template <class numType>
void partSegmented<numType>::create (const mapType<numType> &theMap,
	int partCount, const numType &delta)
{
	// create an object for rounding control
	tRounding<numType> rounding;

	int nCrit = theMap. countCritical ();
	if (partCount < 2 * nCrit + 2)
		throw "Too small partition requested.";
	this -> allocate (partCount);
	(*this) [0] = theMap. leftBound ();
	(*this) [partCount] = theMap. rightBound ();
	int prev = 0;
	for (int i = 0; i < nCrit; ++ i)
	{
		int n = partCount / (nCrit + 1) * (i + 1);
		this -> addCritical (n);
		const numType crit = theMap. criticalPoint (i);
		const numType left = rounding. sub_up (crit, delta);
		const numType right = rounding. add_down (crit, delta);
		if ((left <= (*this) [prev]) ||
			((*this) [partCount] <= right))
		{
			throw "Too large critical neighborhood requested.";
		}
		(*this) [n] = left;
		fillUniform (prev, n);
		(*this) [n + 1] = right;
		prev = n + 1;
	}
	fillUniform (prev, partCount);
	return;
} /* partSegmented::create */


} // namespace unifexp

#endif // _partsegm_h_

/// @}

