/// @addtogroup unifexp
/// @{

/////////////////////////////////////////////////////////////////////////////
///
/// @file parttype.h
///
/// This file contains the definition of an abstract partition type.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2007 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on March 8, 2007. Last revision: November 17, 2013.

#ifndef _parttype_h_
#define _parttype_h_

#include <vector>
#include <string>
#include <algorithm>
#include <string>
#include "chomp/struct/bitfield.h"
#include "chomp/system/textfile.h" // for debugging only
#include "maptype.h"


namespace unifexp {

// --------------------------------------------------
// ----------------- partition type -----------------
// --------------------------------------------------

/// An abstract partition type.
/// All the other partition types should inherit from this one.
template <class numType>
class partType
{
public:
	/// The constructor.
	partType (const std::string &name);

	/// The virtual destructor.
	virtual ~partType ();

	/// Returns the name of the object.
	const std::string &name () const;

	/// Creates a partition based on the given map, the requested
	/// number of elements in the partition, and the width of the
	/// critical neighborhood.
	virtual void create (const mapType<numType> &theMap, int partCount,
		const numType &delta) = 0;

	/// Runs the bisection algorithm to find the maximal element in the
	/// partition table which is less than or equal to the given number.
	/// Uses operator < for comparision, operator [] to extract from
	/// the table. Returns the number of that element or -1 if the number
	/// is below the leftmost limit.
	int find (const numType &number) const;

	/// Modifies the partition slightly in such a way
	/// that the given value becomes one of the bounds
	/// of the partition intervals.
	/// Does not move the bounds of critical intervals.
	/// Throws an exception if it is not possible to modify
	/// the partition, e.g., because the interval is too large.
	/// Returns the number of the bound (index for operator []).
	int include (const numType &number);

	/// Returns the given interval bound from the partition: the left end
	/// of the n-th interval, or right end of the (n-1)-th one.
	/// Does not verify the correctness of the index.
	/// Note that the operator [] which accesses a non-constant object
	/// is protected, so it may be necessary to use a const reference
	/// to access the data of a non-const object with operator [].
	/// In case of any problems, use the function "at" instead (it is
	/// also safer, because it verifies the correctness of the index).
	const numType &operator [] (int n) const;

	/// Returns the given interval bound from the partition:
	/// the left endpoint of the n-th interval, or, equivalently,
	/// the right endendpoint of the (n-1)-th interval (if n > 0).
	/// Verifies the correctness of the index and throws an error message
	/// if n is out of range.
	const numType &at (int n) const;

	/// Returns true if the given interval is part of a neighborhood
	/// of some critical point. Returns false otherwise.
	/// Note that this function does not verify the correctness of n,
	/// so it must be used with caution.
	bool isCritical (int n) const;

	/// Returns the number of the interval which is a neighborhood
	/// of a critical point, or throws an exception if 'n' out of scope.
	/// See the map class for the actual number of critical points.
	int getCritical (int n = 0) const;

protected:
	/// Allocates the table to store the given number of intervals.
	/// It is automatically deallocated by the destructor.
	void allocate (int n);

	/// Adds a critical interval to the list of critical intervals.
	void addCritical (int n);

	/// Returns the given interval bound from the partition: the left end
	/// of the n-th interval, or right end of the (n-1)-th one.
	/// Does not verify the correctness of the index.
	numType &operator [] (int n);

private:
	/// The table with the bounds of partition intervals.
	/// Initially this address is set to 0.
	/// Use the function "allocate" to (re)allocate this table,
	/// and the operator [] and function "at" to access its elements.
	numType *intBounds;

	/// The number of intervals in the partition.
	/// Initialized to 0, set by the function "allocate".
	int intCount;

	/// The list of critical intervals.
	std::vector<int> critical;

	/// A bit field which stores the information about intervals
	/// if they are critical or not.
	chomp::homology::BitField b;

	/// The name of the partition type.
	std::string partName;

	/// Copy constructor is not implemented.
	partType (const partType<numType> &);

	/// Assignment operator is not implemented.
	partType<numType> &operator = (const partType<numType> &);

}; /* class partType */

// --------------------------------------------------

template <class numType>
inline partType<numType>::partType (const std::string &name):
	intBounds (0), intCount (0), partName (name)
{
	return;
} /* partType::partType */

template <class numType>
inline partType<numType>::partType (const partType<numType> &)
{
	throw "Copy constructor not implemented for partitions.";
	return;
} /* partType::partType */

template <class numType>
inline const std::string &partType<numType>::name () const
{
	return partName;
} /* partType::name */

template <class numType>
inline partType<numType> &partType<numType>::operator =
	(const partType<numType> &)
{
	throw "Assignment operator not implemented for partitions.";
	return *this;
} /* partType::operator = */

template <class numType>
inline partType<numType>::~partType ()
{
	if (intBounds)
	{
		delete [] intBounds;
		b. free ();
	}
	return;
} /* partType::~partType */

template <class numType>
inline const numType &partType<numType>::operator [] (int n) const
{
	return intBounds [n];
} /* partType::operator [] */

template <class numType>
inline numType &partType<numType>::operator [] (int n)
{
	return intBounds [n];
} /* partType::operator [] */

template <class numType>
inline const numType &partType<numType>::at (int n) const
{
	if (!intBounds || (n < 0) || (n > intCount))
		throw "Partition element out of range requested.";
	return intBounds [n];
} /* partType::at */

template <class numType>
inline bool partType<numType>::isCritical (int n) const
{
	return !!b. test (n);
//	return (std::find (critical. begin (), critical. end (), n) !=
//		critical. end ());
} /* partType::isCritical */

template <class numType>
inline int partType<numType>::getCritical (int n) const
{
	if ((n < 0) || (n >= static_cast<int> (critical. size ())))
		throw "Critical interval number out of range.";
	return critical [n];
} /* partType::getCritical */

// --------------------------------------------------

template <class numType>
inline int partType<numType>::find (const numType &number) const
{
	if (!intBounds)
		throw "Partition not allocated.";
	if (number < intBounds [0])
		return -1;
	if (number >= intBounds [intCount])
		return intCount;
	int lowerbound = 0;
	int upperbound = intCount + 1;
	while (upperbound - lowerbound > 1)
	{
		int middle = (lowerbound + upperbound) >> 1;
		if (number < intBounds [middle])
			upperbound = middle;
		else
			lowerbound = middle;
	}
	return lowerbound;
} /* partType::find */

template <class numType>
inline int partType<numType>::include (const numType &number)
{
	using chomp::homology::sbug;

	// find an interval in the partition which contains the number
	if (!intBounds)
		throw "Partition not allocated.";
	if (number < intBounds [0])
		throw "Trying to include a number that is too small.";
	if (number == intBounds [0])
		return 0;
	if (number == intBounds [intCount])
		return intCount;
	if (number > intBounds [intCount])
		throw "Trying to include a number that is too big.";
	int index = this -> find (number);

	// if the value agrees with one of the bounds then return its index
	if (intBounds [index] == number)
		return index;
	if ((index < intCount) && (intBounds [index + 1] == number))
		return (index + 1);

	// debug check: the number should be strictly between the bounds
	if ((intBounds [index] >= number) ||
		(intBounds [index + 1] <= number))
	{
		throw "Programmer's error: Included number not in interior.";
	}

	// move one of the bounds
	if (isCritical (index))
	{
		sbug << "DEBUG: number = " << number <<
			", index = " << index << ".\n" <<
			"DEBUG: valueMin = " << intBounds [index] <<
			", valueMax = " << intBounds [index + 1] << ".\n";
		throw "Trying to include a value "
			"between two critical bounds.";
	}
	if (!index || isCritical (index - 1) ||
		(((number - intBounds [index]) >
		(intBounds [index + 1] - number)) &&
		!isCritical (index + 1)))
	{
		++ index;
	}
	intBounds [index] = number;

	return index;
} /* partType::include */

template <class numType>
inline void partType<numType>::addCritical (int n)
{
	critical. push_back (n);
	b. set (n);
	return;
} /* partType::addCritical */

template <class numType>
inline void partType<numType>::allocate (int n)
{
	if (n <= 0)
	{
		throw "A non-positive number of partition intervals "
			"requested.";
	}
	if (intBounds)
	{
		delete [] intBounds;
		b. free ();
	}
	intBounds = new numType [n + 1];
	intCount = n;
	b. allocate (n);
	b. clearall (n);
	critical. clear ();
	return;
} /* partType::allocate */


} // namespace unifexp

#endif // _parttype_h_

/// @}

