/////////////////////////////////////////////////////////////////////////////
///
/// @file quadratic.cpp
///
/// This file contains the main program which processes the command-line
/// arguments and runs the computations.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2006-2015 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on June 11, 2013. Last revision: April 29, 2015.

// NOTE: This program uses the BOOST, MPFR, CAPD and CHomP libraries.


// the first local header file
#include "rounding.h" // (this header must be included first - BOOST)

// other local header files
#include "quadrcoord.h"
#include "quadrworker.h"
#include "maptypes.h"
#include "parttypes.h"

// standard library headers
#include <string>
#include <vector>
#include <algorithm>
#include <exception>

// selected files from the CHomP library
#include "chomp/system/config.h"
#include "chomp/system/textfile.h"
#include "chomp/system/timeused.h"
#include "chomp/multiwork/mw.h"
#include "chomp/system/arg.h"


// --------------------------------------------------
// -------------------- OVERTURE --------------------
// --------------------------------------------------

/// The title of the main program.
const char *title = "\
QuadraticMap, ver. 0.00, April 29, 2015. (C) 2006-2015 by Pawel Pilarczyk.\n\
This is free software. No warranty. Consult 'license.txt' for details.";

/// The help information about the main program.
const char *helpinfo = "\
This is a computer program for rigorous computation of the quantities\n\
that appear in the paper by A. Golmakani, S. Luzzatto, and P. Pilarczyk.\n\
Please, see the paper and the software documentation for details.\n\
Command line arguments (using default values if the argument is omitted):\n\
-a N - the left bound of the parameter interval (obligatory argument),\n\
-a N - the right bound of the parameter interval (obligatory argument),\n\
-n N - the number of subintervals of the parameter range (1),\n\
-r N - restrict the computations to the given subinterval (may be repeated),\n\
-k N - the number of intervals for the uniform expansion estimates (1000),\n\
-k0 N - as -k N, but for obtaining an estimate for delta (1000),\n\
-d N - the radius of the critical neighborhood, a.k.a. delta (0.01),\n\
-part name - the partition type while computing lambda0 (derivative),\n\
-part0 name - the partition type while estimating delta (uniform),\n\
-f name - save text results of the computations to this file,\n\
-l name - load previously computed delta/lambda combinations from the file,\n\
\t(use -e3 -l name to compute these combinations and save/append),\n\
-e N - run experimental computations no. N: 0=none (default), 1=N0 test,\n\
\t2=delta & N0 test, 3=min delta & good lambda test (save data),\n\
-s N - use algorithms for sparse graphs: 0=no, 1=yes, -1=auto (default),\n\
-t N - set the coordinator's walltime to N hours (floating point number),\n\
--select N - select only those subintervals whose numbers divide N (1),\n\
--fraction N - send fractions of intervals, e.g. 100 means intervals of width\n\
\t1/100 and the same starting points (default: 1 = full intervals),\n\
--margin N - set the margin fraction while computing min delta for lambda > 0,\n\
--memcons 0 - don't use the memory conservative algorithm for lambda,\n\
--flush - flush the data written to the results file (slow),\n\
--debug - show debug information on progress of the computations,\n\
--verbose [N] - set higher verbosity of the debug and extra information,\n\
--noflushbug - don't flush the debug stream (faster at high verbosity),\n\
--help - show this brief help information and exit.\n\
==============================================================================\n\
Note: This is software in development. This program is a reduced version\n\
capable of conducting the desired computations, but with limited functionality;\n\
in particular, some of the command line arguments may have no effect.\n\
==============================================================================\n\
Arguments for running distributed computing:\n\
-w computer:port - run a worker and connect to the given coordinator,\n\
-c [port] - run the coordinator and listen at the given port (def: 4721).";

/// Information on where to find more information about the program.
const char *helpmore = "\
For more information ask the author at http://www.pawelpilarczyk.com/.";

/// The control number for network communication. Change this number
/// whenever an incompatible version of this program is created.
const int controlNumber = 42140;


// --------------------------------------------------
// ---------------------- MAIN ----------------------
// --------------------------------------------------

/// The main function of the program.
/// Returns: 0 = Ok, -1 = Error, 1 = Help displayed, 2 = Wrong arguments.
int main (int argc, char *argv [])
{
	using namespace chomp::homology;
	using namespace chomp::multiwork;
	using namespace unifexp;

	// prepare user-configurable data
	const double infinity = 1.7e308;
	double paramMin = -infinity;
	double paramMax = -infinity;
	int paramIntervals = 1;
	int partCount0 = 1000;
	int partCount = 1000;
	double delta = 0.01;
	double deltaPlus = 0.03;
	char *partName0 = const_cast<char *> ("uniform");
	char *partName = const_cast<char *> ("derivative");
	int precision = 1000;
	int port = 0;
	char *work = 0;
	bool test = false;
	bool flushfile = false;
	char *fileName = 0;
	const int maxLambdaNames = 100;
	int nLambdaNames = 0;
	char *lambdaNames [maxLambdaNames];
	int sparseGraph = -1;
	int memoryConservative = 1;
	const int choiceProbesMax = 100;
	int choiceProbesArray [choiceProbesMax];
	int choiceProbesCount = 0;
	const int choicesMax = 100;
	double choicesArray [choicesMax];
	int choicesCount = 0;
	int choiceNum = 0;
	int maxDepth = 0;
	int experimental = 0;
	const int maxComments = 16;
	int nComments = 0;
	char *comments [maxComments];
	double timeLimit = 0;
	int verbosity = 0;
	bool noflushbug = false;
	long fraction = 1;
	const int restrMax = 10000;
	int restrCount = 0;
	int restrArray [restrMax];
	int select = 1;
	double minWidth = 1e-10;
	int maxComp = 0;
	int minN0 = 0;
	int maxN0 = 0;
	double margin = 0;

	// interprete the command-line arguments
	arguments a;
	arg (a, "a", paramMin);
	arg (a, "a", paramMax);
	arg (a, "e", experimental);
	arg (a, "u", choiceProbesArray, choiceProbesCount, choiceProbesMax);
	arg (a, "v", choicesArray, choicesCount, choicesMax);
	arg (a, "V", choiceNum);
	arg (a, "m", maxDepth);
	arg (a, "n", paramIntervals);
	arg (a, "r", restrArray, restrCount, restrMax);
	arg (a, "d", delta);
	arg (a, "D", deltaPlus);
	arg (a, "k0", partCount0);
	arg (a, "k", partCount);
	arg (a, "part0", partName0);
	arg (a, "part", partName);
	arg (a, "prec", precision);
	arg (a, "f", fileName);
	arg (a, "l", lambdaNames, nLambdaNames, maxLambdaNames);
	arg (a, "c", port, 4721);
	arg (a, "w", work);
	arg (a, "s", sparseGraph);
	arg (a, "t", timeLimit);
	arg (a, "-verbose", verbosity, 1000);
	arg (a, "-fraction", fraction);
	arg (a, "-select", select);
	arg (a, "-minwidth", minWidth);
	arg (a, "-maxcomp", maxComp);
	arg (a, "-minN0", minN0);
	arg (a, "-maxN0", maxN0);
	arg (a, "-margin", margin);
	arg (a, "-memcons", memoryConservative);
	arg (a, "-comment", comments, nComments, maxComments);
	argswitch (a, "-noflushbug", noflushbug, true);
	argswitch (a, "-test", test, true);
	argswitch (a, "-flush", flushfile, true);
	arghelp (a);

	argstreamprepare (a);
	int argresult = a. analyze (argc, argv);
	argstreamset ();

	// show the program's main title
	if (argresult >= 0)
		sout << title << '\n';

	// if something was incorrect, show an additional message and exit
	if (argresult < 0)
	{
		sout << "Call with '--help' for help.\n";
		return 2;
	}

	// make sure delta is positive unless it is to be computed
	if ((delta <= 0) || (delta >= deltaPlus))
		argresult = 1;

	// make sure that the parameters were defined
	if (((paramMax == -infinity) || (paramMin > paramMax))
		&& !work && !test)
	{
		argresult = 1;
	}

	// if help requested, show help information
	if (argresult > 0)
	{
		sout << helpinfo << '\n';
		typedef std::vector<std::string> strings;

		// show available partition types
		strings partNames;
		partTypes<double> partitions;
		partitions. getNames (partNames);
		sout << "Available partitions:";
		int pos = 14;
		for (strings::iterator it = partNames. begin ();
			it != partNames. end (); ++ it)
		{
			if (pos + (*it). length () > 70)
			{
				sout << "\n\t";
				pos = 0;
			}
			else if (pos > 0)
				sout << " ";
			sout << (*it);
			pos += (*it). length () + 1;
		}
		sout << ".\n";

		sout << helpmore << '\n';
		return 1;
	}

	// try running the main function and catch an error message if thrown
	try
	{
		// turn off flushing the debug/extra messages if requested to
		if (noflushbug)
			sbug. flush = false;

		// adjust the margin fraction if wrong value was given
		if (margin < 0)
			margin = 0;

		// set an appropriate program time message
		program_time = "Aborted after:";
		program_time = 1;

		// run a test if this option was selected
		if (test)
		{
			auto_mpfr fraction (1000);
			mpfr_set_d (fraction, -10, GMP_RNDN);
			auto_mpfr measure (1000);
			mpfr_exp (measure, fraction, GMP_RNDN);
			char measureStr [1024];
			mpfr_sprintf (measureStr, "%.20RDg",
				static_cast<mpfr_srcptr> (measure));
			sout << "TEST: " << measureStr << ".\n";

			double x (699);
			for (int i = 0; i < 15; ++ i)
			{
				x += 0.1;
				sout << "exp (" << x << ") = ";
				double result (0);
				exp_up (result, x);
				sout << result << ".\n";
			}
		}

		// run as a worker if requested to
		else if (work)
		{
			Worker w (controlNumber, false);
			w. Port (0);
			w. Add (work);
			w. TimeOut (600);
			sout << "Running as a worker...\n";
			int result = w. Work ();
			if (result == mwOk)
				sout << "Work completed successfully.\n";
			else
				sout << "Could not work - probably "
					"an error occurred.\n";
		}

		// run as a coordinator or run the work locally otherwise
		else
		{
			// create a coordinator
			std::vector<int> choiceProbesVector
				(choiceProbesCount);
			std::copy (choiceProbesArray,
				choiceProbesArray + choiceProbesCount,
				choiceProbesVector. begin ());
			std::vector<double> choicesVector (choicesCount);
			std::copy (choicesArray, choicesArray + choicesCount,
				choicesVector. begin ());
			std::vector<int> restrVector (restrCount);
			std::copy (restrArray, restrArray + restrCount,
				restrVector. begin ());
			Coordinator c (controlNumber, fileName, flushfile,
				paramMin, paramMax,
				paramIntervals, restrVector,
				lambdaNames, nLambdaNames, delta, deltaPlus,
				partName, partName0, partCount, partCount0,
				sparseGraph, memoryConservative,
				precision, choiceProbesVector, choicesVector,
				choiceNum, maxDepth, experimental,
				timeLimit, verbosity, fraction, select,
				minWidth, maxComp, minN0, maxN0, margin);
			c. Port (port);

			// run as a listening coordinator if requested to
			if (port)
			{
				sout << "Running as a coordinator...\n";
				int result = c. Coordinate ();
				if (result == mwOk)
					sout << "Completed successfully.\n";
				else
					sout << "Could not coordinate - "
						"an error occurred.\n";
			}

			// run the computations locally, otherwise
			else
			{
				Worker w (controlNumber, true);
				sout << "Running the computations...\n";
				int result = c. Coordinate (&w);
				if (result == mwOk)
					sout << "Completed successfully.\n";
				else
					sout << "Could not work - probably "
						"an error occurred.\n";
			}
		}

		// set an appropriate program time message
		program_time = "Total time used:";

		// finalize
		return 0;
	}
	catch (const char *msg)
	{
		sout << "ERROR: " << msg << '\n';
		return -1;
	}
	catch (const std::exception &e)
	{
		sout << "ERROR: " << e. what () << '\n';
		return -1;
	}
	catch (...)
	{
		sout << "ABORT: An unknown error occurred.\n";
		return -1;
	}
} /* main */

