/////////////////////////////////////////////////////////////////////////////
///
/// @file quadrcomp.h
///
/// The main computation routines for the quadratic map.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2006-2014 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on June 12, 2013. Last revision: April 29, 2015.


#ifndef _quadrcomp_h_
#define _quadrcomp_h_

// local header files
#include "uniflambda.h"
#include "numdouble.h"
using namespace numdouble;

// standard library headers
#include <string>
#include <vector>
#include <algorithm>
#include <exception>
#include <iostream>
#include <new>
#include <memory>
//#define NDEBUG
#include <assert.h>

// selected header files from the CHomP library
#include "chomp/system/config.h"
#include "chomp/system/textfile.h"
#include "chomp/system/timeused.h"


// --------------------------------------------------
// --------------------- TESTS ----------------------
// --------------------------------------------------

//#define CHEATING


// --------------------------------------------------
// ------------------- verbosity --------------------
// --------------------------------------------------

/// Minimal verbosity levels requied to show additional types of information.
namespace Verbosity
{

const int gammaPrim (30);
const int N1 (20);
const int Pk (30);
const int pMu (30);
const int muStar (20);
const int KMu (30);
const int LMu (30);
const int conditions (15);
const int condFailDetails (26);
const int condFailure (14);
const int alphaTilde (20);
const int NHat (20);
const int muDelta (20);
const int muDeltaPlus (20);
const int formalConsts (20);
const int compConsts (10);
const int omega (15);
const int N0 (15);
const int lnQuot (25);
const int choice (13);
const int etaChoice (12);
const int choiceFailure (14);
const int choiceTime (13);
const int Omega (6);
const int recursionWarnings (8);
const int recursionInfo (9);
const int badWidthOmegaN0 (9);
const int delta (4);
const int deltaPlus (4);
const int omegaFull (4);
const int unifExp (5);
const int lambda (4);
const int Gamma (30);
const int recursive (3);
const int eta (4);
const int etaFinal (1);
const int minWidth (4);
const int maxN0 (4);
const int cheatKMu (24);
const int findLambda (25);

} /* namespace Verbosity */


/// Conducts a single computation with the given number.
/// Appends the results to the provided vector using "push_back".
template <class MapType, class PartType>
inline void runComputation (double delta, double /*deltaPlus*/,
	double paramLeft, double paramRight,
	const MapType &theMap, PartType &thePart, PartType &thePart0,
	int partCount, int partCount0,
	int sparseGraph, bool memoryConservative,
	int precision, const std::vector<int> &choiceProbesVector,
	const std::vector<double> &choicesVector, int choiceNum,
	int maxDepth, bool knownLambda, double lambdaCycles0,
	double lambdaStart0, double lambdaEnd0, int experimental,
	std::vector<double> &results, int verbosity,
	double minWidthFraction, int maxComp, int minN0, int maxN0,
	double margin)
{
	using chomp::homology::sout;
	using chomp::homology::sbug;
	using unifexp::IntervalType;
	using unifexp::resetRounding;
	const int hiprec = 12;
	const int stdprec = std::cout. precision ();

	if (verbosity >= Verbosity::delta)
		sbug << "delta = " << delta << ".\n";
	double deltaPlus (0);
	mul_up (deltaPlus, delta, 3);
	if (verbosity >= Verbosity::deltaPlus)
		sbug << "deltaPlus = " << deltaPlus << ".\n";
	if (verbosity >= Verbosity::omegaFull)
	{
		sbug << "omegaFull = [" << std::setprecision (hiprec) <<
			paramLeft << ", " << paramRight <<
			std::setprecision (stdprec) << "].\n";
	}

	// experimental computation no. 3: compute delta, then a good lambda
//	if (experimental == 3)
	{
		// prepare the parameters for this type of computation
		const int deltaSteps = 20;

		// find min delta for which lambda can be found positive
		// using a small partition; use the provided delta
		// as an upper bound for the allowed delta
		double deltaGood (delta);
		double lambdaCyclesGood (0);
		double lambdaStartGood (0);
		double lambdaEndGood (0);
		double deltaBad (0);
		double deltaTry (deltaGood);
		double deltaPP (0);
	//	sbug << "Finding a good delta...\n";
		for (int step = 0; step < deltaSteps; ++ step)
		{
			double lambdaCycles (0);
			double lambdaStart (0);
			double lambdaEnd (0);
			{
				using chomp::homology::outputstream;
				std::auto_ptr<outputstream::mute> mout
					((verbosity >= Verbosity::findLambda)
					? 0 : new outputstream::mute (sout));
				std::auto_ptr<outputstream::mute> mbug
					((verbosity >= Verbosity::findLambda)
					? 0 : new outputstream::mute (sbug));
				mul_up (deltaPP, deltaTry, 3);
				deltaPP *= 2;
				findLambda (deltaTry, deltaPP, theMap,
					thePart0, partCount0,
					lambdaCycles, lambdaStart, lambdaEnd,
					sparseGraph, memoryConservative);
			}
			sbug << " [" << step << "] delta = " << deltaTry <<
				", lambdas: " << lambdaCycles << ", " <<
				lambdaStart << ", " << lambdaEnd << ".\n";
			double lambdaMin (lambdaCycles);
			// (min3 (lambdaCycles,
			//	lambdaStart, lambdaEnd));
			if ((step == 0) && (lambdaMin <= 0))
				throw "Negative lambda for maximal delta.";
			if (lambdaMin > 0)
			{
				deltaGood = deltaTry;
				lambdaCyclesGood = lambdaCycles;
				lambdaStartGood = lambdaStart;
				lambdaEndGood = lambdaEnd;
			}
			else
			{
				deltaBad = deltaTry;
			}
			deltaTry = (deltaGood + deltaBad) / 2;
		}
		sbug << "deltaMin = " << deltaGood << ".\n";
		sbug << "lambdaMin: " << lambdaCyclesGood << ", " <<
			lambdaStartGood << ", " << lambdaEndGood << ".\n";
		double deltaBetter (deltaGood);
		if (margin > 0)
		{
			deltaBetter *= 1.0 + margin;
			sbug << "deltaMin1 = " << deltaBetter << ".\n";
		}

		// find a better lambda by using a fine partition
		double lambdaCycles (0);
		double lambdaStart (0);
		double lambdaEnd (0);
		{
			using chomp::homology::outputstream;
			std::auto_ptr<outputstream::mute> mout
				((verbosity >= Verbosity::findLambda)
				? 0 : new outputstream::mute (sout));
			std::auto_ptr<outputstream::mute> mbug
				((verbosity >= Verbosity::findLambda)
				? 0 : new outputstream::mute (sbug));
			mul_up (deltaPP, deltaBetter, 3);
			deltaPP *= 2;
			findLambda (deltaBetter, deltaPP, theMap, thePart,
				partCount, lambdaCycles,
				lambdaStart, lambdaEnd,
				sparseGraph, memoryConservative);
		}
		sbug << "lambdas: " << lambdaCycles << ", " <<
				lambdaStart << ", " << lambdaEnd << ".\n";

		// return the results of the computation
		results. push_back (deltaGood);
		results. push_back (lambdaCyclesGood);
		results. push_back (lambdaStartGood);
		results. push_back (lambdaEndGood);
		results. push_back (deltaBetter);
		results. push_back (lambdaCycles);
		results. push_back (lambdaStart);
		results. push_back (lambdaEnd);
		return;
	}

	return;
} /* runComputation */


#endif // _quadrcomp_h_

