/////////////////////////////////////////////////////////////////////////////
///
/// @file quadrcoord.h
///
/// This file contains the definition of the class "coordinator"
/// which coordinates series of computations.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2006-2013 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on June 12, 2013. Last revision: April 27, 2015.


#ifndef _quadrcoord_h_
#define _quadrcoord_h_

#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>

#include "chomp/system/config.h"
#include "chomp/system/textfile.h"
#include "chomp/system/timeused.h"
#include "chomp/multiwork/mw.h"
//#include "chomp/struct/multitab.h"

#include "datavector.h"
#include "expdata.h"
#include "mpfrsupp.h"


// --------------------------------------------------
// ------------- the coordinator class --------------
// --------------------------------------------------

/// An object of this class coordinates a series of computations.
/// This class inherits from the mwCoordinator class
/// to allow network distribution of the work.
class Coordinator: public chomp::multiwork::mwCoordinator
{
public:
	/// The complete constructor of a coordinator.
	Coordinator (int _controlNumber,
		const char *_fileName, bool _flushfile,
		double _paramMin, double _paramMax,
		int _paramIntervals, const std::vector<int> &_restrVector,
		char * const _lambdaNames [], int _nLambdaNames,
		double _delta, double _deltaPlus,
		const char *_partName, const char *_partName0,
		int _partCount, int _partCount0,
		int _sparseGraph, bool _memoryConservative, int _precision,
		const std::vector<int> &_choiceProbesVector,
		const std::vector<double> &_choicesVector,
		int _choiceNum, int _maxDepth, int _experimental,
		double _timeLimit, int _verbosity,
		long _fraction, int _select, double _minWidth,
		int _maxComp, int _minN0, int _maxN0, double _margin);

	/// The destructor of the coordinator.
	~Coordinator ();

private:
	/// A function for preparing data by a coordinator.
	int Prepare (chomp::multiwork::mwData &data);

	/// A function for accepting results by a coordinator.
	int Accept (chomp::multiwork::mwData &data);

	/// A function for taking rejected data by a coordinator.
	int Reject (chomp::multiwork::mwData &data);

	// --- Parameters of Computations ---

	/// The control number to recognize compatible workers.
	int controlNumber;

	/// Should the data pieces be processed in the inverted order
	/// (starting from the last one)?
	bool invOrder;

	/// A file to append the results to.
	std::ofstream f;

	/// Should the results file be flushed after every write?
	bool flushfile;

	/// The minimal value of the parameter 'a' to investigate.
	double paramMin;

	/// The maximal value of the parameter 'a' to investigate.
	double paramMax;

	/// Expansion data structure.
	std::vector<ExpansionData *> expData;

	/// The minimal value of delta if fixed.
	double delta;

	/// The maximal value of delta if fixed.
	double deltaPlus;

	/// The partition type name.
	std::string partName;

	/// The partition type name for estimating delta.
	std::string partName0;

	/// The minimal size of the partition.
	int partCount;

	/// The size of the partition for estimating delta.
	int partCount0;

	/// The numbers of probes in each variable.
	std::vector<int> choiceProbesVector;

	/// The user's choices vector.
	std::vector<double> choicesVector;

	/// The requested number of the choice (if positive).
	int choiceNum;

	/// The maximal allowed recursion depth (if positive).
	int maxDepth;

	// --- Current Data of Computations ---

	/// Reads the previously computed results from the input file.
	/// Returns true if successful, false if no results file exists.
	bool readResults (const char *fileName);

	/// Opens a file with the results for appending or creates the file.
	/// If this is a new file then some introductory info is written.
	void openResults (const char *fileName, bool newFile);

	/// Calculates the given bound of the current interval.
	double intervalBound (int cur, int all);

	/// The total number of subintervals to process (either real numbers
	/// of offsets in the restriction vector).
	int maxCurrent;

	/// The currently processed subinterval.
	int current;

	/// The total number of subintervals of parameters at the top level.
	int paramIntervals;

	/// The specific numbers of elements to which the computation
	/// should be restricted.
	std::vector<int> restrVector;

	/// Should a sparse graph algorithm be used? [0=no, 1=yes, -1=auto]
	int sparseGraph;

	/// Should the memory conservative graph algorithm be used
	/// for computing lambda?
	bool memoryConservative;

	/// The binary precision of high-precision computations.
	int precision;

	/// The number of experimental computation.
	int experimental;

	/// The wall time for the computations (if nonzero) in seconds.
	double wallTime;

	/// The longest computation time encountered so far (in seconds).
	double longestTime;

	/// The coordinator's running time.
	chomp::homology::timeused runTime;

	/// The low precision for writing real numbers to files.
	int loprec;

	/// The high precision for writing real numbers to files.
	int hiprec;

	/// The verbosity of output while running the computations.
	int verbosity;

	/// The fraction of full intervals of parameters to use (1 = full).
	long fraction;

	/// Selection criterion: Use only these subintervals whose
	/// consecutive numbers divide the provided number (1 = use all).
	int select;

	/// The minimal allowed width of subintervals for which recursion
	/// can be continued. This number is given as a fraction of the
	/// original subinterval of parameters. Must be positive.
	double minWidth;

	/// The maximal number of computations that are allowed to be done
	/// for each parameter subinterval (0 = no limits).
	int maxComp;

	/// The minimal value of N0 for which computations can be launched.
	int minN0;

	/// The maximal value of N0 for which the recursion is continued
	/// (0 = no limits).
	int maxN0;

	/// The safety margin fraction that is added while computing
	/// the minimal delta for which lambda > 0.
	double margin;

	/// The cumulative measure of the interval of stochastic parameters.
	auto_mpfr measure;

}; /* class Coordinator */

// --------------------------------------------------

inline bool Coordinator::readResults (const char *fileName)
{
	using namespace chomp::homology;

	// ignore a file with no or empty name
	if (!fileName && !*fileName)
		return false;

	// open the file for reading
	std::ifstream in (fileName);
	if (!in)
		return false;

	// process all the lines with the results and mark which are read
	ignorecomments (in);
	while (in. peek () != EOF)
	{
		if (in. peek () != '*')
		{
			ignoreline (in);
			ignorecomments (in);
			continue;
		}
		in. get ();
		ignorecomments (in);

		// TODO!

		ignoreline (in);
		ignorecomments (in);
	}
	return true;
} /* Coordinator::readResults */

inline void Coordinator::openResults (const char *fileName, bool newFile)
{
	using namespace chomp::homology;

	// ignore a file with no or empty name
	if (!fileName && !*fileName)
		return;

	// open the results file for appending
	f. open (fileName, std::ios::out | std::ios::app);
	f << "; Started on " << currenttime ();
	if (!f)
		fileerror (fileName);

	// indicate the meaning of the data in each line
	if (!newFile)
		return;
	if (experimental == 1)
	{
		f << "; parMin parMax N0\n";
	}
	else if (experimental == 2)
	{
		f << "; parMin parMax delta N0\n";
	}
	else if (experimental == 3)
	{
		f << "; parMin parMax deltaMin "
			"lambdaCyclesMin lambdaStartMin lambdaEndMin "
			"delta lambdaCycles lambdaStart lambdaEnd compTime\n";
	}
	else if (experimental)
	{
	}
	else
	{
		f << "; parMin parMax eta compTime\n";
	}

	return;
} /* Coordinator::openResults */

inline double Coordinator::intervalBound (int cur, int all)
{
	// return the trivial answer if this is the case
	if (cur == 0)
		return paramMin;
	if (cur == all)
		return paramMax;

	// calculate the GCD of 'cur' and 'all' (code based on Wikipedia)
	int a (all);
	int b (cur);
	while (b)
	{
		int t = b;
		b = a % b;
		a = t;
	}
	// DEBUG check if this is correct
	if (true)
	{
		if ((cur % a) || (all % a))
			throw "Wrong GCD calculated.";
	}

	// reduce the fraction cur/all
	cur /= a;
	all /= a;

	// calculate the interval bound and return it
	return (paramMin + (cur * (paramMax - paramMin) / all));
} /* Coordinator::intervalBound */

inline Coordinator::Coordinator (int _controlNumber,
	const char *_fileName, bool _flushfile,
	double _paramMin, double _paramMax,
	int _paramIntervals, const std::vector<int> &_restrVector,
	char * const _lambdaNames [], int _nLambdaNames,
	double _delta, double _deltaPlus,
	const char *_partName, const char *_partName0,
	int _partCount, int _partCount0, int _sparseGraph,
	bool _memoryConservative, int _precision,
	const std::vector<int> &_choiceProbesVector,
	const std::vector<double> &_choicesVector,
	int _choiceNum, int _maxDepth, int _experimental,
	double _timeLimit, int _verbosity, long _fraction, int _select,
	double _minWidth, int _maxComp, int _minN0, int _maxN0,
	double _margin):
	controlNumber (_controlNumber), invOrder (true),
	flushfile (_flushfile),
	paramMin (_paramMin), paramMax (_paramMax),
	delta (_delta), deltaPlus (_deltaPlus),
	partName (_partName), partName0 (_partName0),
	partCount (_partCount), partCount0 (_partCount0),
	choiceProbesVector (_choiceProbesVector),
	choicesVector (_choicesVector), choiceNum (_choiceNum),
	maxDepth (_maxDepth),
	maxCurrent (_restrVector. empty () ? _paramIntervals :
	_restrVector. size ()),
	current (invOrder ? (maxCurrent - 1) : 0),
	paramIntervals (_paramIntervals), restrVector (_restrVector),
	sparseGraph (_sparseGraph), memoryConservative (_memoryConservative),
	precision (_precision),
	experimental (_experimental),
	wallTime (static_cast<int> (_timeLimit * 3600)), longestTime (0),
	loprec (6), hiprec (15), verbosity (_verbosity),
	fraction (_fraction), select (_select), minWidth (_minWidth),
	maxComp (_maxComp), minN0 (_minN0), maxN0 (_maxN0), margin (_margin),
	measure (precision)
{
	using namespace chomp::homology;

	// check if the numbers of the items in the restriction vector are ok
	for (std::vector<int>::const_iterator it = restrVector. begin ();
		it != restrVector. end (); ++ it)
	{
		if (((*it) < 0) || ((*it) >= paramIntervals))
			throw "Restriction item number out of range.";
	}

	// refrain from displaying messages about coordinator's running time
	runTime = 0;

	// process and open the results file if its file name is given
	if (_fileName && *_fileName)
	{
		// read the previously computed results of the computations
		bool resultsRead = this -> readResults (_fileName);
		if (!resultsRead)
		{
			sout << "Note: Could not open the results file. "
				"Any previous results will be ignored.\n";
		}

		// open the results file for appending
		openResults (_fileName, !resultsRead);
	}

	// otherwise, warn the user
	else
		sout << "Warning: Results will not be saved to a file.\n";

	// should a file be opened for writing?
	bool lambdaWriting = (experimental == 3);
	if (lambdaWriting && !_nLambdaNames)
		throw "Wrong number of expansion data files provided.";

	// read the expansion data structures from files
	bool writingReady = false;
	for (int n = 0; n < _nLambdaNames; ++ n)
	{
		// remember if this is the last file in the series
		// (which is important in the case of writing)
		bool lastFile = (n == _nLambdaNames - 1);

		// get the successive file name and check if it is defined
		const char *_lambdaName = _lambdaNames [n];
		if (_lambdaName && *_lambdaName)
		{
			// create a new data structure
			ExpansionData *expNew = new ExpansionData;

			// if writing, the data structure in the file
			// must be compatible with the current data
			if (lambdaWriting && lastFile)
			{
				expNew -> initialize (paramMin, paramMax,
					paramIntervals);
			}

			// read the data from previous computations
			sout << "Reading expansion data from '" <<
				_lambdaName << "'... ";
			bool headerRead = false;
			std::ifstream in;
			in. open (_lambdaName, std::ios::binary);

			// if the file could not be opened then skip it,
			// unless the file should be created for writing
			if (!in)
			{
				sout << "Can't open the file.\n";
				if (!(lambdaWriting && lastFile))
					continue;
			}

			// if the file is empty then ignore it,
			// unless the file should be used for writing
			else if (in. peek () == EOF)
			{
				sout << "The file is empty.\n";
				if (!(lambdaWriting && lastFile))
					continue;
			}

			// if the file was opened successfully then read it
			else
			{
				if (expNew -> readHeader (in) < 0)
				{
					sout << "Incompatible data.\n";
					continue;
				}
				headerRead = true;
				int_t count = expNew -> readData (in);
				sout << count << " data pieces loaded.\n";
				in. close ();
			}

			// if the results should be written to this file
			if (lambdaWriting && lastFile)
			{
				// reopen the previously read file,
				// or create a new one
				if (headerRead)
				{
					sout << "Appending expansion data "
						"to the existing file.\n";
				}
				else
				{
					sout << "Creating a new expansion "
						"data file.\n";
				}
				std::ofstream *out = new std::ofstream ();
				out -> open (_lambdaName, std::ios::binary |
					std::ios::app);
				if (!*out)
				{
					throw "Can't create or re-open "
						"the expansion data file.";
				}

				// write the header if necessary
				if (!headerRead)
					expNew -> writeHeader (*out);

				// keep the stream for writing the data
				expNew -> setStream (out);
				writingReady = true;
			}

			// add the expansion data file to the collection
			expData. push_back (expNew);
		}
	}

	if (lambdaWriting && !writingReady)
		throw "Couldn't open a expansion data file for writing.";

	// make a correction to the fraction if necessary
	if (fraction <= 0)
		fraction = 1;

	// initialize the cumulative measure
	mpfr_set_zero (measure, 0);

	return;
} /* Coordinator::Coordinator */

inline Coordinator::~Coordinator ()
{
	using namespace chomp::homology;

	// write the information about the total measure gathered so far
	if (mpfr_sgn (measure) > 0)
	{
		std::string measureString (mpfr2stringDown (measure));
		sout << "A lower bound for the measure of stochastic "
			"parameters found in this computation:\nm = " <<
			measureString << ".\n";
		f << "M = " << measureString << ".\n";
	}

	// close the text results file
	f << "; Finished on " << currenttime ();
	f. close ();

	// close the expansion data streams and free the allocated variables
	for (std::vector<ExpansionData *>::iterator it = expData. begin ();
		it != expData. end (); ++ it)
	{
		ExpansionData &expDataBlock = *(*it);
		std::ostream *expDataStream = expDataBlock. getStream ();
		if (expDataStream)
		{
			expDataBlock. setStream (0);
			delete expDataStream;
		}
	}

	return;
} /* Coordinator::~Coordinator */

inline int Coordinator::Prepare (chomp::multiwork::mwData &data)
{
	using namespace chomp::homology;
	using namespace chomp::multiwork;

	// reset the data to prepare
	data. Reset ();

	// if the wall time is getting close then prepare no more data
	static bool finalizing = false;
	if (finalizing)
		return mwNoData;
	if (wallTime && (static_cast<double> (runTime) +
		2 * longestTime > wallTime))
	{
		finalizing = true;
		static bool displayed = false;
		if (!displayed)
		{
			sout << "The walltime of " << wallTime << " sec "
				"is getting close; longest comp = " <<
				longestTime << " sec.\n"
				"=== Finalizing the computations... ===\n";
			displayed = true;
		}
		return mwNoData;
	}

	// skip all the current data that have already been processed before
	// or are not selected using the division criterion
	int currentInterval = -1;
	for (; (invOrder && (current >= 0)) ||
		(!invOrder && (current < maxCurrent));
		current += (invOrder ? -1 : 1))
	{
		currentInterval = restrVector. empty () ?
			current : restrVector [current];
		if ((select > 1) && (currentInterval % select))
			continue;
		if (experimental == 1)
		{
		}
		else if (experimental == 2)
		{
		}
		else if (experimental == 3)
		{
			double delta (0);
			double lambdaCycles (0);
			double lambdaStart (0);
			double lambdaEnd (0);
			expData [expData. size () - 1] -> get
				(currentInterval, delta,
				lambdaCycles, lambdaStart, lambdaEnd);
			if (delta != 0)
				continue;
		}
		else if (experimental)
		{
		}
		else
		{
		}
		break;
	}

	// if there is no more data to send then say this
	if ((invOrder && (current < 0)) ||
		(!invOrder && (current >= maxCurrent)))
	{
		return mwNoData;
	}

	// calculate an appropriate parameter interval to send
	double paramLeft = intervalBound (currentInterval, paramIntervals);
	double paramRight = (fraction > 1) ? intervalBound
		(currentInterval * fraction + 1, paramIntervals * fraction) :
		intervalBound (currentInterval + 1, paramIntervals);
//	double paramLeft = (currentInterval == 0) ? paramMin : (paramMin +
//		(currentInterval * (paramMax - paramMin) / paramIntervals));
//	double paramRight = (fraction > 1) ?
//		(paramMin + (currentInterval * fraction + 1) *
//		(paramMax - paramMin) / (paramIntervals * fraction)) :
//		((currentInterval == paramIntervals - 1) ?
//		paramMax : (paramMin + (currentInterval + 1) *
//		(paramMax - paramMin) / paramIntervals));

	// check if optimal delta and lambda is known for this interval
	double delta0 (0);
	double lambdaCycles0 (0);
	double lambdaStart0 (0);
	double lambdaEnd0 (0);
	bool knownLambda = findBestExpansion (expData, paramLeft, paramRight,
		delta0, lambdaCycles0, lambdaStart0, lambdaEnd0);

	// prepare data to send
	data << currentInterval;
	data << partName;
	data << partName0;
	data << partCount;
	data << partCount0;
	data << paramLeft;
	data << paramRight;
	data << (knownLambda ? delta0 : delta);
	data << (knownLambda ? (3 * delta0) : deltaPlus);
	data << knownLambda;
	data << lambdaCycles0;
	data << lambdaStart0;
	data << lambdaEnd0;
	data << sparseGraph;
	data << memoryConservative;
	data << precision;
	data << choiceProbesVector;
	data << choicesVector;
	data << choiceNum;
	data << maxDepth;
	data << experimental;
	data << loprec;
	data << hiprec;
	data << verbosity;
	data << minWidth;
	data << maxComp;
	data << minN0;
	data << maxN0;
	data << margin;
	data << controlNumber;

	// show a message on what has been done
	sout << "+ " << current << ".\n";

	// take the current item number to use the next time
	current += (invOrder ? -1 : 1);

	return mwOk;
} /* Coordinator::Prepare */

inline int Coordinator::Accept (chomp::multiwork::mwData &data)
{
	using namespace chomp::homology;
	using namespace chomp::multiwork;

	// decode the data
	int cur = 0;
	data >> cur;
	double pMin = 0;
	data >> pMin;
	double pMax = 0;
	data >> pMax;
	std::vector<double> results;
	data >> results;
	double compTime = 0;
	data >> compTime;
	int ctrl = 0;
	data >> ctrl;
	if (ctrl != controlNumber)
	{
		f << "! Wrong data received.\n";
		throw "Wrong data received from a worker.";
	}

	// update the longest encountered computation time
	if (longestTime < compTime)
		longestTime = compTime;

	// interprete the array of results obtained from the worker
	double eta (0);
	double deltaGood (0);
	double deltaBetter (0);
	double lambdaCyclesGood (0);
	double lambdaStartGood (0);
	double lambdaEndGood (0);
	double lambdaCycles (0);
	double lambdaStart (0);
	double lambdaEnd (0);
	if (experimental == 1)
	{
		if (results. size () > 0)
			eta = results [0];
	}
	else if (experimental == 2)
	{
		if (results. size () > 0)
			eta = results [0];
	}
	else if (experimental == 3)
	{
		if (results. size () >= 3)
		{
			deltaGood = results [0];
			lambdaCyclesGood = results [1];
			lambdaStartGood = results [2];
			lambdaEndGood = results [3];
			deltaBetter = results [4];
			lambdaCycles = results [5];
			lambdaStart = results [6];
			lambdaEnd = results [7];
			expData [expData. size () - 1] -> set
				(cur, deltaBetter, lambdaCycles,
				lambdaStart, lambdaEnd);
		}
	}
	else if (experimental)
	{
	}
	else
	{
		if (results. size () > 0)
			eta = results [results. size () - 1];
	}

	// say what has been received and save the data to the output file
	sout << "* " << cur << " [" << std::setprecision (hiprec) <<
		pMin << "," << pMax << std::setprecision (loprec) << "]";
	f << std::setprecision (hiprec) << pMin << " " << pMax <<
		std::setprecision (loprec);
	if (experimental == 1)
	{
		sout << " N0=" << eta;
		f << " " << eta;
	}
	else if (experimental == 2)
	{
		int N0 = static_cast<int> (eta);
		double delta = (eta - N0) * 4;
		sout << " delta=" << delta << " N0=" << N0 <<
			" (" << compTime << "s)";
	//	f << std::setprecision (hiprec) << delta <<
	//		std::setprecision (loprec) << " " << N0;
		f << " " << delta << " " << N0;
	}
	else if (experimental == 3)
	{
	//	sout << " minDelta: " << minDelta;
		f << " " << deltaGood << " " << lambdaCyclesGood << " " <<
			lambdaStartGood << " " << lambdaEndGood << " " <<
			deltaBetter << " " << lambdaCycles << " " <<
			lambdaStart << " " << lambdaEnd << " " << compTime;
	}
	else if (experimental)
	{
	}
	else
	{
		sout << " eta=" << eta << " (" << compTime << "s)";
		f << " " << std::setprecision (hiprec) << eta <<
			std::setprecision (loprec) << " " << compTime;

		// add the fraction of the interval length to the measure
		if (eta > 0)
		{
			auto_mpfr left (precision);
			mpfr_set_d (left, pMin, GMP_RNDU);
			auto_mpfr right (precision);
			mpfr_set_d (right, pMax, GMP_RNDD);
			auto_mpfr width (precision);
			mpfr_sub (width, right, left, GMP_RNDD);
			auto_mpfr etaWidth (precision);
			mpfr_mul_d (etaWidth, width, eta, GMP_RNDD);
			auto_mpfr measurePrev;
			mpfr_swap (measure, measurePrev);
			mpfr_add (measure, measurePrev, etaWidth, GMP_RNDD);
		}
	}
	sout << "\n";
	f << "\n";

	// flush the data file if requested to
	if (flushfile)
		f << std::flush;

	return mwOk;
} /* Coordinator::Accept */

inline int Coordinator::Reject (chomp::multiwork::mwData &data)
{
	using namespace chomp::homology;
	using namespace chomp::multiwork;

	// decode the number
	int cur = 0;
	data >> cur;

	if (experimental == 3)
	{
		expData [expData. size () - 1] -> set (cur, -1, -1, -1, -1);
	}

	// say that data has been rejected
	sout << "Data " << cur << " rejected.\n";

	return mwOk;
} /* Coordinator::Reject */

#endif // _quadrcoord_h_

