/////////////////////////////////////////////////////////////////////////////
///
/// @file quadrworker.h
///
/// This file contains the definition of the worker class
/// whose objects process single chunks of data.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2006-2013 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on June 12, 2013. Last revision: April 27, 2015.


#ifndef _quadrworker_h_
#define _quadrworker_h_

#include <vector>

#include "chomp/system/config.h"
#include "chomp/system/textfile.h"
#include "chomp/system/timeused.h"
#include "chomp/multiwork/mw.h"

#include "maptypes.h"
#include "parttypes.h"
#include "quadrcomp.h"
#include "datavector.h"


// --------------------------------------------------
// ---------------- the worker class -----------------
// --------------------------------------------------

/// The worker class that processes single chunks of data.
class Worker: public chomp::multiwork::mwWorker
{
public:
	/// The default constructor.
	Worker (int _controlNumber, bool _local = false);

private:
	/// The control number to recognize a compatible coordinator.
	int controlNumber;

	/// Is the work being done locally? That is, in the same process
	/// as the coordinator? If so, then the global variables are not
	/// reset, such as pointbase, because this might affect the
	/// coordinator's data.
	bool local;

	/// A function for the initialization of a worker.
	int Initialize (chomp::multiwork::mwData &data);

	/// A function for processing a piece of data by a worker.
	int Process (chomp::multiwork::mwData &data);

private:
	/// An object which contains all the partition objects
	/// to choose from.
	unifexp::partTypes<double> partitions;

}; /* class Worker */

// --------------------------------------------------

inline Worker::Worker (int _controlNumber, bool _local):
	controlNumber (_controlNumber), local (_local)
{
	return;
} /* Worker::Worker */

// --------------------------------------------------

inline int Worker::Initialize (chomp::multiwork::mwData &/*data*/)
{
	// ignore the received data and say it was Ok
	return chomp::multiwork::mwOk;
} /* Worker::Initialize */

inline int Worker::Process (chomp::multiwork::mwData &data)
{
	using namespace chomp::homology;
	using namespace chomp::multiwork;

	// remember the time used so far
	double timeStamp = static_cast<double> (program_time);

	// decode the number of the data pack to process
	int current = 0;
	data >> current;

	// say which data pack is being processed
	sout << "- " << current << ".\n";

	// the map object to use
	const char *mapName = "minus-quadratic-intv";
	unifexp::mapTypes<double> maps;
	unifexp::mapType<double> *theMap = maps. get (mapName);
	if (!theMap)
	{
		sout << "! Cannot find the map '" << mapName << "'.\n";
		data. Reset ();
		return mwReject;
	}

	// the partition object to use
	std::string partName;
	data >> partName;
	unifexp::partType<double> *thePart = partitions. get (partName);
	if (!thePart)
	{
		sout << "! Cannot find the partition '" <<
			partName << "'.\n";
		data. Reset ();
		return mwReject;
	}

	// the partition object to use for estimating delta
	std::string partName0;
	data >> partName0;
	unifexp::partType<double> *thePart0 = partitions. get (partName0);
	if (!thePart0)
	{
		sout << "! Cannot find the partition '" <<
			partName0 << "'.\n";
		data. Reset ();
		return mwReject;
	}

	// the size of the partition
	int partCount = 0;
	data >> partCount;

	// the size of the partition for estimating delta
	int partCount0 = 0;
	data >> partCount0;

	// the parameter range
	double paramLeft = 0;
	data >> paramLeft;
	double paramRight = 0;
	data >> paramRight;
	theMap -> setParam (paramLeft, paramRight);

	// delta and delta+
	double delta = 0;
	data >> delta;
	double deltaPlus = 0;
	data >> deltaPlus;

	// lambda (if known)
	bool knownLambda = false;
	data >> knownLambda;
	double lambdaCycles (0);
	data >> lambdaCycles;
	double lambdaStart (0);
	data >> lambdaStart;
	double lambdaEnd (0);
	data >> lambdaEnd;

	// should sparse graph algorithms be used?
	int sparseGraph = -1;
	data >> sparseGraph;

	// should the memory conservative (but slower) algorithm be used?
	bool memoryConservative = true;
	data >> memoryConservative;

	// the binary precision of high-precision computations
	int precision = 0;
	data >> precision;

	// the vector of user's choices of values of configurable parameters
	std::vector<int> choiceProbesVector;
	data >> choiceProbesVector;

	// the vector of user's choices of values of configurable parameters
	std::vector<double> choicesVector;
	data >> choicesVector;

	// the user's choice (by number)
	int choiceNum = 0;
	data >> choiceNum;

	// the maximal recursion depth
	int maxDepth = 0;
	data >> maxDepth;

	// determine the number of experimental computations to run if any
	int experimental = 0;
	data >> experimental;

	// the precision for writing real numbers to the screen and log files
	int loprec = 6;
	data >> loprec;
	int hiprec = 6;
	data >> hiprec;

	// the verbosity of debugging output
	int verbosity = 0;
	data >> verbosity;

	// the minimal width of intervals for recursion (as a fraction)
	double minWidth (0);
	data >> minWidth;

	// max number of computations that can be launched (0 = no limits)
	int maxComp (0);
	data >> maxComp;

	// min N0 for which the computations can be done
	int minN0 (0);
	data >> minN0;

	// max N0 for which the recursion can be continued
	int maxN0 (0);
	data >> maxN0;

	/// savety margin while computing min delta for which lambda > 0
	double margin (0);
	data >> margin;

	// the ending control number
	int ctrl = 0;
	data >> ctrl;
	if (ctrl != controlNumber)
	{
		sout << "! Data incomplete. Rejecting it.\n";
		data. Reset ();
		return mwReject;
	}

	// remember the default precision and set a good one for 'sbug'
	int outprec = std::cout. precision ();
//	sbug << std::setprecision (hiprec);
//	sout << std::setprecision (hiprec);

	// run one set of computations
	int result = 0;
	std::string rejectInfo;
	std::vector<double> results;
	try
	{
		// compute all the necessary constants,
		// check the restrictions, and finally compute eta
		// (or run the experimental computation)
		runComputation (delta, deltaPlus,
			paramLeft, paramRight, *theMap, *thePart, *thePart0,
			partCount, partCount0,
			sparseGraph, memoryConservative,
			precision, choiceProbesVector, choicesVector,
			choiceNum, maxDepth, knownLambda,
			lambdaCycles, lambdaStart, lambdaEnd,
			experimental, results, verbosity,
			minWidth, maxComp, minN0, maxN0, margin);
	}
	catch (const char *msg)
	{
		rejectInfo = msg;
		result = -1;
	}
	catch (const std::exception &e)
	{
		rejectInfo = e. what ();
		result = -1;
	}
	catch (...)
	{
		rejectInfo = "An unknown error occurred.";
		result = -1;
	}

	// restore the defaulut precision of the two major output streams
	sbug << std::setprecision (outprec);
	sout << std::setprecision (outprec);

	// determine the computation time and show it
	double compTime (static_cast<double> (program_time) - timeStamp);

	// say what has been computed
	sout << ((result < 0) ? "! " : "= ") << current <<
		" (" << compTime << " sec).\n";

	// if the result is wrong then reject this piece of data
	if (result < 0)
	{
		data. Reset ();
		sout << "REJECT: " << rejectInfo << '\n';
		return mwReject;
	}

	// send back the data containing the result of the processing
	data. Reset ();
	data << current;
	data << paramLeft;
	data << paramRight;
	data << results;
	data << compTime;
	data << controlNumber;

	// return the result
	return mwOk;
} /* Worker::Process */

#endif // _quadrworker_h_

